@@ -37,21 +37,13 @@ typedef struct {
 
     PyObject *wrapped;
     PyObject *buffer;
-    /* NOTE: this used to be a PostgreSQL encoding: changed in 2.3.2 to be a
-     * Python codec name. I don't expect there has been any user for this
-     * object other than adapting str/unicode, so I don't expect client code
-     * broken for this reason. */
-    char     *encoding;
 
-    PyObject *conn;
+    connectionObject *conn;
+
+    const char *encoding;
+
 } qstringObject;
 
-/* functions exported to psycopgmodule.c */
-
-HIDDEN PyObject *psyco_QuotedString(PyObject *module, PyObject *args);
-#define psyco_QuotedString_doc \
-    "QuotedString(str, enc) -> new quoted string"
-
 #ifdef __cplusplus
 }
 #endif
