@@ -34,20 +34,41 @@
 
 #include <string.h>
 
-/* Mapping from isolation level name to value exposed by Python.
- *
- * Note: ordering matters: to get a valid pre-PG 8 level from one not valid,
- * we increase a pointer in this list by one position. */
-const IsolationLevel conn_isolevels[] = {
-    {"",                    ISOLATION_LEVEL_AUTOCOMMIT},
-    {"read uncommitted",    ISOLATION_LEVEL_READ_UNCOMMITTED},
-    {"read committed",      ISOLATION_LEVEL_READ_COMMITTED},
-    {"repeatable read",     ISOLATION_LEVEL_REPEATABLE_READ},
-    {"serializable",        ISOLATION_LEVEL_SERIALIZABLE},
-    {"default",            -1},     /* never to be found on the server */
-    { NULL }
+/* String indexes match the ISOLATION_LEVEL_* consts */
+const char *srv_isolevels[] = {
+    NULL, /* autocommit */
+    "READ COMMITTED",
+    "REPEATABLE READ",
+    "SERIALIZABLE",
+    "READ UNCOMMITTED",
+    "default"       /* only to set GUC, not for BEGIN */
 };
 
+/* Read only false, true */
+const char *srv_readonly[] = {
+    " READ WRITE",
+    " READ ONLY",
+    ""      /* default */
+};
+
+/* Deferrable false, true */
+const char *srv_deferrable[] = {
+    " NOT DEFERRABLE",
+    " DEFERRABLE",
+    ""      /* default */
+};
+
+/* On/Off/Default GUC states
+ */
+const char *srv_state_guc[] = {
+    "off",
+    "on",
+    "default"
+};
+
+
+const int SRV_STATE_UNCHANGED = -1;
+
 
 /* Return a new "string" from a char* from the database.
  *
@@ -58,12 +79,75 @@ const IsolationLevel conn_isolevels[] = {
 PyObject *
 conn_text_from_chars(connectionObject *self, const char *str)
 {
-#if PY_MAJOR_VERSION < 3
-        return PyString_FromString(str);
-#else
-        const char *codec = self ? self->codec : "ascii";
-        return PyUnicode_Decode(str, strlen(str), codec, "replace");
-#endif
+    return psycopg_text_from_chars_safe(str, -1, self ? self->pydecoder : NULL);
+}
+
+
+/* Encode an unicode object into a bytes object in the connection encoding.
+ *
+ * If no connection or encoding is available, default to utf8
+ */
+PyObject *
+conn_encode(connectionObject *self, PyObject *u)
+{
+    PyObject *t = NULL;
+    PyObject *rv = NULL;
+
+    if (!(self && self->pyencoder)) {
+        rv = PyUnicode_AsUTF8String(u);
+        goto exit;
+    }
+
+    if (!(t = PyObject_CallFunctionObjArgs(self->pyencoder, u, NULL))) {
+        goto exit;
+    }
+
+    if (!(rv = PyTuple_GetItem(t, 0))) { goto exit; }
+    Py_INCREF(rv);
+
+exit:
+    Py_XDECREF(t);
+
+    return rv;
+}
+
+
+/* decode a c string into a Python unicode in the connection encoding
+ *
+ * len can be < 0: in this case it will be calculated
+ *
+ * If no connection or encoding is available, default to utf8
+ */
+PyObject *
+conn_decode(connectionObject *self, const char *str, Py_ssize_t len)
+{
+    PyObject *b = NULL;
+    PyObject *t = NULL;
+    PyObject *rv = NULL;
+
+    if (len < 0) { len = strlen(str); }
+
+    if (self) {
+        if (self->cdecoder) {
+            return self->cdecoder(str, len, NULL);
+        }
+        else if (self->pydecoder) {
+            if (!(b = Bytes_FromStringAndSize(str, len))) { goto exit; }
+            if (!(t = PyObject_CallFunctionObjArgs(self->pydecoder, b, NULL))) {
+                goto exit;
+            }
+            rv = PyTuple_GetItem(t, 0);
+            Py_XINCREF(rv);
+        }
+    }
+    else {
+        return PyUnicode_FromStringAndSize(str, len);
+    }
+
+exit:
+    Py_XDECREF(t);
+    Py_XDECREF(b);
+    return rv;
 }
 
 /* conn_notice_callback - process notices */
@@ -87,13 +171,20 @@ conn_notice_callback(void *args, const char *message)
         /* Discard the notice in case of failed allocation. */
         return;
     }
+    notice->next = NULL;
     notice->message = strdup(message);
     if (NULL == notice->message) {
         free(notice);
         return;
     }
-    notice->next = self->notice_pending;
-    self->notice_pending = notice;
+
+    if (NULL == self->last_notice) {
+        self->notice_pending = self->last_notice = notice;
+    }
+    else {
+        self->last_notice->next = notice;
+        self->last_notice = notice;
+    }
 }
 
 /* Expose the notices received as Python objects.
@@ -104,44 +195,60 @@ void
 conn_notice_process(connectionObject *self)
 {
     struct connectionObject_notice *notice;
-    Py_ssize_t nnotices;
+    PyObject *msg = NULL;
+    PyObject *tmp = NULL;
+    static PyObject *append;
 
     if (NULL == self->notice_pending) {
         return;
     }
 
-    notice =  self->notice_pending;
-    nnotices = PyList_GET_SIZE(self->notice_list);
+    if (!append) {
+        if (!(append = Text_FromUTF8("append"))) {
+            goto error;
+        }
+    }
 
+    notice = self->notice_pending;
     while (notice != NULL) {
-        PyObject *msg;
-        msg = conn_text_from_chars(self, notice->message);
         Dprintf("conn_notice_process: %s", notice->message);
 
-        /* Respect the order in which notices were produced,
-           because in notice_list they are reversed (see ticket #9) */
-        if (msg) {
-            PyList_Insert(self->notice_list, nnotices, msg);
-            Py_DECREF(msg);
-        }
-        else {
-            /* We don't really have a way to report errors, so gulp it.
-             * The function should only fail for out of memory, so we are
-             * likely going to die anyway. */
-            PyErr_Clear();
+        if (!(msg = conn_text_from_chars(self, notice->message))) { goto error; }
+
+        if (!(tmp = PyObject_CallMethodObjArgs(
+                self->notice_list, append, msg, NULL))) {
+
+            goto error;
         }
 
+        Py_DECREF(tmp); tmp = NULL;
+        Py_DECREF(msg); msg = NULL;
+
         notice = notice->next;
     }
 
     /* Remove the oldest item if the queue is getting too long. */
-    nnotices = PyList_GET_SIZE(self->notice_list);
-    if (nnotices > CONN_NOTICES_LIMIT) {
-        PySequence_DelSlice(self->notice_list,
-            0, nnotices - CONN_NOTICES_LIMIT);
+    if (PyList_Check(self->notice_list)) {
+        Py_ssize_t nnotices;
+        nnotices = PyList_GET_SIZE(self->notice_list);
+        if (nnotices > CONN_NOTICES_LIMIT) {
+            if (-1 == PySequence_DelSlice(self->notice_list,
+                    0, nnotices - CONN_NOTICES_LIMIT)) {
+                PyErr_Clear();
+            }
+        }
     }
 
     conn_notice_clean(self);
+    return;
+
+error:
+    Py_XDECREF(tmp);
+    Py_XDECREF(msg);
+    conn_notice_clean(self);
+
+    /* TODO: the caller doesn't expects errors from us */
+    PyErr_Clear();
 }
 
 void
@@ -154,11 +261,11 @@ conn_notice_clean(connectionObject *self)
     while (notice != NULL) {
         tmp = notice;
         notice = notice->next;
-        free((void*)tmp->message);
+        free(tmp->message);
         free(tmp);
     }
 
-    self->notice_pending = NULL;
+    self->last_notice = self->notice_pending = NULL;
 }
 
 
@@ -173,6 +280,15 @@ conn_notifies_process(connectionObject *self)
     PGnotify *pgn = NULL;
     PyObject *notify = NULL;
     PyObject *pid = NULL, *channel = NULL, *payload = NULL;
+    PyObject *tmp = NULL;
+
+    static PyObject *append;
+
+    if (!append) {
+        if (!(append = Text_FromUTF8("append"))) {
+            goto error;
+        }
+    }
 
     while ((pgn = PQnotifies(self->pgconn)) != NULL) {
 
@@ -183,7 +299,7 @@ conn_notifies_process(connectionObject *self)
         if (!(channel = conn_text_from_chars(self, pgn->relname))) { goto error; }
         if (!(payload = conn_text_from_chars(self, pgn->extra))) { goto error; }
 
-        if (!(notify = PyObject_CallFunctionObjArgs((PyObject *)&NotifyType,
+        if (!(notify = PyObject_CallFunctionObjArgs((PyObject *)&notifyType,
                 pid, channel, payload, NULL))) {
             goto error;
         }
@@ -192,7 +308,11 @@ conn_notifies_process(connectionObject *self)
         Py_DECREF(channel); channel = NULL;
         Py_DECREF(payload); payload = NULL;
 
-        PyList_Append(self->notifies, (PyObject *)notify);
+        if (!(tmp = PyObject_CallMethodObjArgs(
+                self->notifies, append, notify, NULL))) {
+            goto error;
+        }
+        Py_DECREF(tmp); tmp = NULL;
 
         Py_DECREF(notify); notify = NULL;
         PQfreemem(pgn); pgn = NULL;
@@ -201,6 +321,7 @@ conn_notifies_process(connectionObject *self)
 
 error:
     if (pgn) { PQfreemem(pgn); }
+    Py_XDECREF(tmp);
     Py_XDECREF(notify);
     Py_XDECREF(pid);
     Py_XDECREF(channel);
@@ -226,7 +347,7 @@ conn_get_standard_conforming_strings(PGconn *pgconn)
      * The presence of the 'standard_conforming_strings' parameter
      * means that the server _accepts_ the E'' quote.
      *
-     * If the paramer is off, the PQescapeByteaConn returns
+     * If the parameter is off, the PQescapeByteaConn returns
      * backslash escaped strings (e.g. '\001' -> "\\001"),
      * so the E'' quotes are required to avoid warnings
      * if 'escape_string_warning' is set.
@@ -284,142 +405,168 @@ exit:
     return rv;
 }
 
-/* Convert a PostgreSQL encoding to a Python codec.
- *
- * Set 'codec' to a new copy of the codec name allocated on the Python heap.
- * Return 0 in case of success, else -1 and set an exception.
- *
- * 'enc' should be already normalized (uppercase, no - or _).
+/* set fast access functions according to the currently selected encoding
  */
-RAISES_NEG static int
-conn_encoding_to_codec(const char *enc, char **codec)
+static void
+conn_set_fast_codec(connectionObject *self)
 {
-    char *tmp;
-    Py_ssize_t size;
-    PyObject *pyenc = NULL;
-    int rv = -1;
+    Dprintf("conn_set_fast_codec: encoding=%s", self->encoding);
 
-    /* Find the Py codec name from the PG encoding */
-    if (!(pyenc = PyDict_GetItemString(psycoEncodings, enc))) {
+    if (0 == strcmp(self->encoding, "UTF8")) {
+        Dprintf("conn_set_fast_codec: PyUnicode_DecodeUTF8");
+        self->cdecoder = PyUnicode_DecodeUTF8;
+        return;
+    }
+
+    if (0 == strcmp(self->encoding, "LATIN1")) {
+        Dprintf("conn_set_fast_codec: PyUnicode_DecodeLatin1");
+        self->cdecoder = PyUnicode_DecodeLatin1;
+        return;
+    }
+
+    Dprintf("conn_set_fast_codec: no fast codec");
+    self->cdecoder = NULL;
+}
+
+
+/* Return the Python encoding from a PostgreSQL encoding.
+ *
+ * Optionally return the clean version of the postgres encoding too
+ */
+PyObject *
+conn_pgenc_to_pyenc(const char *encoding, char **clean_encoding)
+{
+    char *pgenc = NULL;
+    PyObject *rv = NULL;
+
+    if (0 > clear_encoding_name(encoding, &pgenc)) { goto exit; }
+    if (!(rv = PyDict_GetItemString(psycoEncodings, pgenc))) {
         PyErr_Format(OperationalError,
-            "no Python codec for client encoding '%s'", enc);
+            "no Python encoding for PostgreSQL encoding '%s'", pgenc);
         goto exit;
     }
+    Py_INCREF(rv);
 
-    /* Convert the codec in a bytes string to extract the c string. */
-    Py_INCREF(pyenc);
-    if (!(pyenc = psycopg_ensure_bytes(pyenc))) {
-        goto exit;
+    if (clean_encoding) {
+        *clean_encoding = pgenc;
     }
-
-    if (-1 == Bytes_AsStringAndSize(pyenc, &tmp, &size)) {
-        goto exit;
+    else {
+        PyMem_Free(pgenc);
     }
 
-    /* have our own copy of the python codec name */
-    rv = psycopg_strdup(codec, tmp, size);
-
 exit:
-    Py_XDECREF(pyenc);
     return rv;
 }
 
-/* Read the client encoding from the connection.
+/* Convert a Postgres encoding into Python encoding and decoding functions.
  *
- * Store the encoding in the pgconn->encoding field and the name of the
- * matching python codec in codec. The buffers are allocated on the Python
- * heap.
+ * Set clean_encoding to a clean version of the Postgres encoding name
+ * and pyenc and pydec to python codec functions.
  *
- * Return 0 on success, else nonzero.
+ * Return 0 on success, else -1 and set an exception.
+ */
+RAISES_NEG static int
+conn_get_python_codec(const char *encoding,
+    char **clean_encoding, PyObject **pyenc, PyObject **pydec)
+{
+    int rv = -1;
+    char *pgenc = NULL;
+    PyObject *encname = NULL;
+    PyObject *enc_tmp = NULL, *dec_tmp = NULL;
+
+    /* get the Python name of the encoding as a C string */
+    if (!(encname = conn_pgenc_to_pyenc(encoding, &pgenc))) { goto exit; }
+    if (!(encname = psycopg_ensure_bytes(encname))) { goto exit; }
+
+    /* Look up the codec functions */
+    if (!(enc_tmp = PyCodec_Encoder(Bytes_AS_STRING(encname)))) { goto exit; }
+    if (!(dec_tmp = PyCodec_Decoder(Bytes_AS_STRING(encname)))) { goto exit; }
+
+    /* success */
+    *pyenc = enc_tmp; enc_tmp = NULL;
+    *pydec = dec_tmp; dec_tmp = NULL;
+    *clean_encoding = pgenc; pgenc = NULL;
+    rv = 0;
+
+exit:
+    Py_XDECREF(enc_tmp);
+    Py_XDECREF(dec_tmp);
+    Py_XDECREF(encname);
+    PyMem_Free(pgenc);
+
+    return rv;
+}
+
+
+/* Store the encoding in the pgconn->encoding field and set the other related
+ * encoding fields in the connection structure.
+ *
+ * Return 0 on success, else -1 and set an exception.
+ */
+RAISES_NEG static int
+conn_store_encoding(connectionObject *self, const char *encoding)
+{
+    int rv = -1;
+    char *pgenc = NULL;
+    PyObject *enc_tmp = NULL, *dec_tmp = NULL;
+
+    if (0 > conn_get_python_codec(encoding, &pgenc, &enc_tmp, &dec_tmp)) {
+        goto exit;
+    }
+
+    /* Good, success: store the encoding/codec in the connection. */
+    {
+        char *tmp = self->encoding;
+        self->encoding = pgenc;
+        PyMem_Free(tmp);
+        pgenc = NULL;
+    }
+
+    Py_CLEAR(self->pyencoder);
+    self->pyencoder = enc_tmp;
+    enc_tmp = NULL;
+
+    Py_CLEAR(self->pydecoder);
+    self->pydecoder = dec_tmp;
+    dec_tmp = NULL;
+
+    conn_set_fast_codec(self);
+
+    rv = 0;
+
+exit:
+    Py_XDECREF(enc_tmp);
+    Py_XDECREF(dec_tmp);
+    PyMem_Free(pgenc);
+    return rv;
+}
+
+
+/* Read the client encoding from the backend and store it in the connection.
+ *
+ * Return 0 on success, else -1.
  */
 RAISES_NEG static int
 conn_read_encoding(connectionObject *self, PGconn *pgconn)
 {
-    char *enc = NULL, *codec = NULL;
-    const char *tmp;
+    const char *encoding;
     int rv = -1;
 
-    tmp = PQparameterStatus(pgconn, "client_encoding");
-    Dprintf("conn_connect: client encoding: %s", tmp ? tmp : "(none)");
-    if (!tmp) {
+    encoding = PQparameterStatus(pgconn, "client_encoding");
+    Dprintf("conn_connect: client encoding: %s", encoding ? encoding : "(none)");
+    if (!encoding) {
         PyErr_SetString(OperationalError,
             "server didn't return client encoding");
         goto exit;
     }
 
-    if (0 > clear_encoding_name(tmp, &enc)) {
+    if (0 > conn_store_encoding(self, encoding)) {
         goto exit;
     }
 
-    /* Look for this encoding in Python codecs. */
-    if (0 > conn_encoding_to_codec(enc, &codec)) {
-        goto exit;
-    }
-
-    /* Good, success: store the encoding/codec in the connection. */
-    PyMem_Free(self->encoding);
-    self->encoding = enc;
-    enc = NULL;
-
-    PyMem_Free(self->codec);
-    self->codec = codec;
-    codec = NULL;
-
     rv = 0;
 
 exit:
-    PyMem_Free(enc);
-    PyMem_Free(codec);
-    return rv;
-}
-
-
-RAISES_NEG int
-conn_get_isolation_level(connectionObject *self)
-{
-    PGresult *pgres = NULL;
-    char *error = NULL;
-    int rv = -1;
-    char *lname;
-    const IsolationLevel *level;
-
-    /* this may get called by async connections too: here's your result */
-    if (self->autocommit) {
-        return 0;
-    }
-
-    Py_BEGIN_ALLOW_THREADS;
-    pthread_mutex_lock(&self->lock);
-
-    if (!(lname = pq_get_guc_locked(self, "default_transaction_isolation",
-            &pgres, &error, &_save))) {
-        goto endlock;
-    }
-
-    /* find the value for the requested isolation level */
-    level = conn_isolevels;
-    while ((++level)->name) {
-        if (0 == strcasecmp(level->name, lname)) {
-            rv = level->value;
-            break;
-        }
-    }
-    if (-1 == rv) {
-        error = malloc(256);
-        PyOS_snprintf(error, 256,
-            "unexpected isolation level: '%s'", lname);
-    }
-
-    free(lname);
-
-endlock:
-    pthread_mutex_unlock(&self->lock);
-    Py_END_ALLOW_THREADS;
-
-    if (rv < 0) {
-        pq_complete_error(self, &pgres, &error);
-    }
-
     return rv;
 }
 
@@ -439,10 +586,41 @@ conn_get_server_version(PGconn *pgconn)
     return (int)PQserverVersion(pgconn);
 }
 
-PGcancel *
-conn_get_cancel(PGconn *pgconn)
+/* set up the cancel key of the connection.
+ * On success return 0, else set an exception and return -1
+ */
+RAISES_NEG static int
+conn_setup_cancel(connectionObject *self, PGconn *pgconn)
 {
-    return PQgetCancel(pgconn);
+    if (self->cancel) {
+        PQfreeCancel(self->cancel);
+    }
+
+    if (!(self->cancel = PQgetCancel(self->pgconn))) {
+        PyErr_SetString(OperationalError, "can't get cancellation key");
+        return -1;
+    }
+
+    return 0;
+}
+
+/* Return 1 if the "replication" keyword is set in the DSN, 0 otherwise */
+static int
+dsn_has_replication(char *pgdsn)
+{
+    int ret = 0;
+    PQconninfoOption *connopts, *ptr;
+
+    connopts = PQconninfoParse(pgdsn, NULL);
+
+    for(ptr = connopts; ptr->keyword != NULL; ptr++) {
+      if(strcmp(ptr->keyword, "replication") == 0 && ptr->val != NULL)
+        ret = 1;
+    }
+
+    PQconninfoFree(connopts);
+
+    return ret;
 }
 
 
@@ -473,34 +651,29 @@ conn_setup(connectionObject *self, PGconn *pgconn)
 {
     PGresult *pgres = NULL;
     char *error = NULL;
+    int rv = -1;
 
     self->equote = conn_get_standard_conforming_strings(pgconn);
     self->server_version = conn_get_server_version(pgconn);
     self->protocol = conn_get_protocol_version(self->pgconn);
     if (3 != self->protocol) {
         PyErr_SetString(InterfaceError, "only protocol 3 supported");
-        return -1;
+        goto exit;
     }
 
     if (0 > conn_read_encoding(self, pgconn)) {
-        return -1;
+        goto exit;
     }
 
-    self->cancel = conn_get_cancel(self->pgconn);
-    if (self->cancel == NULL) {
-        PyErr_SetString(OperationalError, "can't get cancellation key");
-        return -1;
+    if (0 > conn_setup_cancel(self, pgconn)) {
+        goto exit;
     }
 
     Py_BEGIN_ALLOW_THREADS;
     pthread_mutex_lock(&self->lock);
     Py_BLOCK_THREADS;
 
-    if (psyco_green() && (0 > pq_set_non_blocking(self, 1))) {
-        return -1;
-    }
-
-    if (!conn_is_datestyle_ok(self->pgconn)) {
+    if (!dsn_has_replication(self->dsn) && !conn_is_datestyle_ok(self->pgconn)) {
         int res;
         Py_UNBLOCK_THREADS;
         res = pq_set_guc_locked(self, "datestyle", "ISO",
@@ -508,18 +681,26 @@ conn_setup(connectionObject *self, PGconn *pgconn)
         Py_BLOCK_THREADS;
         if (res < 0) {
             pq_complete_error(self, &pgres, &error);
-            return -1;
+            goto unlock;
         }
     }
 
     /* for reset */
     self->autocommit = 0;
+    self->isolevel = ISOLATION_LEVEL_DEFAULT;
+    self->readonly = STATE_DEFAULT;
+    self->deferrable = STATE_DEFAULT;
 
+    /* success */
+    rv = 0;
+
+unlock:
     Py_UNBLOCK_THREADS;
     pthread_mutex_unlock(&self->lock);
     Py_END_ALLOW_THREADS;
 
-    return 0;
+exit:
+    return rv;
 }
 
 /* conn_connect - execute a connection to the database */
@@ -563,6 +744,9 @@ _conn_sync_connect(connectionObject *self)
 
     /* if the connection is green, wait to finish connection */
     if (green) {
+        if (0 > pq_set_non_blocking(self, 1)) {
+            return -1;
+        }
         if (0 != psyco_wait(self)) {
             return -1;
         }
@@ -604,6 +788,11 @@ _conn_async_connect(connectionObject *self)
 
     PQsetNoticeProcessor(pgconn, conn_notice_callback, (void*)self);
 
+    /* Set the connection to nonblocking now. */
+    if (pq_set_non_blocking(self, 1) != 0) {
+        return -1;
+    }
+
     /* The connection will be completed banging on poll():
      * First with _conn_poll_connecting() that will finish connection,
      * then with _conn_poll_setup_async() that will do the same job
@@ -615,14 +804,23 @@ _conn_async_connect(connectionObject *self)
 int
 conn_connect(connectionObject *self, long int async)
 {
-   if (async == 1) {
+    int rv;
+
+    if (async == 1) {
       Dprintf("con_connect: connecting in ASYNC mode");
-      return _conn_async_connect(self);
+      rv = _conn_async_connect(self);
     }
     else {
       Dprintf("con_connect: connecting in SYNC mode");
-      return _conn_sync_connect(self);
+      rv = _conn_sync_connect(self);
     }
+
+    if (rv != 0) {
+        /* connection failed, so let's close ourselves */
+        self->closed = 2;
+    }
+
+    return rv;
 }
 
 
@@ -632,6 +830,7 @@ static int
 _conn_poll_connecting(connectionObject *self)
 {
     int res = PSYCO_POLL_ERROR;
+    const char *msg;
 
     Dprintf("conn_poll: poll connecting");
     switch (PQconnectPoll(self->pgconn)) {
@@ -646,7 +845,11 @@ _conn_poll_connecting(connectionObject *self)
         break;
     case PGRES_POLLING_FAILED:
     case PGRES_POLLING_ACTIVE:
-        PyErr_SetString(OperationalError, "asynchronous connection failed");
+        msg = PQerrorMessage(self->pgconn);
+        if (!(msg && *msg)) {
+            msg = "asynchronous connection failed";
+        }
+        PyErr_SetString(OperationalError, msg);
         res = PSYCO_POLL_ERROR;
         break;
     }
@@ -773,11 +976,6 @@ _conn_poll_setup_async(connectionObject *self)
 
     switch (self->status) {
     case CONN_STATUS_CONNECTING:
-        /* Set the connection to nonblocking now. */
-        if (pq_set_non_blocking(self, 1) != 0) {
-            break;
-        }
-
         self->equote = conn_get_standard_conforming_strings(self->pgconn);
         self->protocol = conn_get_protocol_version(self->pgconn);
         self->server_version = conn_get_server_version(self->pgconn);
@@ -788,10 +986,8 @@ _conn_poll_setup_async(connectionObject *self)
         if (0 > conn_read_encoding(self, self->pgconn)) {
             break;
         }
-        self->cancel = conn_get_cancel(self->pgconn);
-        if (self->cancel == NULL) {
-            PyErr_SetString(OperationalError, "can't get cancellation key");
-            break;
+        if (0 > conn_setup_cancel(self, self->pgconn)) {
+            return -1;
         }
 
         /* asynchronous connections always use isolation level 0, the user is
@@ -801,8 +997,11 @@ _conn_poll_setup_async(connectionObject *self)
         self->autocommit = 1;
 
         /* If the datestyle is ISO or anything else good,
-         * we can skip the CONN_STATUS_DATESTYLE step. */
-        if (!conn_is_datestyle_ok(self->pgconn)) {
+         * we can skip the CONN_STATUS_DATESTYLE step.
+         * Note that we cannot change the datestyle on a replication
+         * connection.
+         */
+        if (!dsn_has_replication(self->dsn) && !conn_is_datestyle_ok(self->pgconn)) {
             Dprintf("conn_poll: status -> CONN_STATUS_DATESTYLE");
             self->status = CONN_STATUS_DATESTYLE;
             if (0 == pq_send_query(self, psyco_datestyle)) {
@@ -890,7 +1089,7 @@ conn_poll(connectionObject *self)
             }
 
             curs = (cursorObject *)py_curs;
-            IFCLEARPGRES(curs->pgres);
+            CLEARPGRES(curs->pgres);
             curs->pgres = pq_get_last_result(self);
 
             /* fetch the tuples (if there are any) and build the result. We
@@ -918,7 +1117,8 @@ conn_poll(connectionObject *self)
 void
 conn_close(connectionObject *self)
 {
-    if (self->closed) {
+    /* a connection with closed == 2 still requires cleanup */
+    if (self->closed == 1) {
         return;
     }
 
@@ -936,7 +1136,7 @@ conn_close(connectionObject *self)
 
 void conn_close_locked(connectionObject *self)
 {
-    if (self->closed) {
+    if (self->closed == 1) {
         return;
     }
 
@@ -957,8 +1157,6 @@ void conn_close_locked(connectionObject *self)
         PQfinish(self->pgconn);
         self->pgconn = NULL;
         Dprintf("conn_close: PQfinish called");
-        PQfreeCancel(self->cancel);
-        self->cancel = NULL;
     }
 }
 
@@ -984,177 +1182,138 @@ conn_rollback(connectionObject *self)
     return res;
 }
 
+
+/* Change the state of the session */
 RAISES_NEG int
-conn_set_session(connectionObject *self,
-        const char *isolevel, const char *readonly, const char *deferrable,
-        int autocommit)
+conn_set_session(connectionObject *self, int autocommit,
+        int isolevel, int readonly, int deferrable)
 {
+    int rv = -1;
     PGresult *pgres = NULL;
     char *error = NULL;
-    int res = -1;
+    int want_autocommit = autocommit == SRV_STATE_UNCHANGED ?
+        self->autocommit : autocommit;
 
-    Py_BEGIN_ALLOW_THREADS;
-    pthread_mutex_lock(&self->lock);
-
-    if (isolevel) {
-        Dprintf("conn_set_session: setting isolation to %s", isolevel);
-        if ((res = pq_set_guc_locked(self,
-                "default_transaction_isolation", isolevel,
-                &pgres, &error, &_save))) {
-            goto endlock;
-        }
+    if (deferrable != SRV_STATE_UNCHANGED && self->server_version < 90100) {
+        PyErr_SetString(ProgrammingError,
+            "the 'deferrable' setting is only available"
+            " from PostgreSQL 9.1");
+        goto exit;
     }
 
-    if (readonly) {
-        Dprintf("conn_set_session: setting read only to %s", readonly);
-        if ((res = pq_set_guc_locked(self,
-                "default_transaction_read_only", readonly,
-                &pgres, &error, &_save))) {
-            goto endlock;
-        }
-    }
-
-    if (deferrable) {
-        Dprintf("conn_set_session: setting deferrable to %s", deferrable);
-        if ((res = pq_set_guc_locked(self,
-                "default_transaction_deferrable", deferrable,
-                &pgres, &error, &_save))) {
-            goto endlock;
-        }
-    }
-
-    if (self->autocommit != autocommit) {
-        Dprintf("conn_set_session: setting autocommit to %d", autocommit);
-        self->autocommit = autocommit;
-    }
-
-    res = 0;
-
-endlock:
-    pthread_mutex_unlock(&self->lock);
-    Py_END_ALLOW_THREADS;
-
-    if (res < 0) {
-        pq_complete_error(self, &pgres, &error);
-    }
-
-    return res;
-}
-
-int
-conn_set_autocommit(connectionObject *self, int value)
-{
-    Py_BEGIN_ALLOW_THREADS;
-    pthread_mutex_lock(&self->lock);
-
-    self->autocommit = value;
-
-    pthread_mutex_unlock(&self->lock);
-    Py_END_ALLOW_THREADS;
-
-    return 0;
-}
-
-/* conn_switch_isolation_level - switch isolation level on the connection */
-
-RAISES_NEG int
-conn_switch_isolation_level(connectionObject *self, int level)
-{
-    PGresult *pgres = NULL;
-    char *error = NULL;
-    int curr_level;
-    int ret = -1;
-
-    /* use only supported levels on older PG versions */
+    /* Promote an isolation level to one of the levels supported by the server */
     if (self->server_version < 80000) {
-        if (level == ISOLATION_LEVEL_READ_UNCOMMITTED)
-            level = ISOLATION_LEVEL_READ_COMMITTED;
-        else if (level == ISOLATION_LEVEL_REPEATABLE_READ)
-            level = ISOLATION_LEVEL_SERIALIZABLE;
+        if (isolevel == ISOLATION_LEVEL_READ_UNCOMMITTED) {
+            isolevel = ISOLATION_LEVEL_READ_COMMITTED;
+        }
+        else if (isolevel == ISOLATION_LEVEL_REPEATABLE_READ) {
+            isolevel = ISOLATION_LEVEL_SERIALIZABLE;
+        }
     }
 
-    if (-1 == (curr_level = conn_get_isolation_level(self))) {
-        return -1;
-    }
-
-    if (curr_level == level) {
-        /* no need to change level */
-        return 0;
-    }
-
-    /* Emulate the previous semantic of set_isolation_level() using the
-     * functions currently available. */
-
     Py_BEGIN_ALLOW_THREADS;
     pthread_mutex_lock(&self->lock);
 
-    /* terminate the current transaction if any */
-    if ((ret = pq_abort_locked(self, &pgres, &error, &_save))) {
-        goto endlock;
-    }
-
-    if (level == 0) {
-        if ((ret = pq_set_guc_locked(self,
-                "default_transaction_isolation", "default",
-                &pgres, &error, &_save))) {
-            goto endlock;
-        }
-        self->autocommit = 1;
-    }
-    else {
-        /* find the name of the requested level */
-        const IsolationLevel *isolevel = conn_isolevels;
-        while ((++isolevel)->name) {
-            if (level == isolevel->value) {
-                break;
+    if (want_autocommit) {
+        /* we are or are going in autocommit state, so no BEGIN will be issued:
+         * configure the session with the characteristics requested */
+        if (isolevel != SRV_STATE_UNCHANGED) {
+            if (0 > pq_set_guc_locked(self,
+                    "default_transaction_isolation", srv_isolevels[isolevel],
+                    &pgres, &error, &_save)) {
+                goto endlock;
             }
         }
-        if (!isolevel->name) {
-            ret = -1;
-            error = strdup("bad isolation level value");
-            goto endlock;
+        if (readonly != SRV_STATE_UNCHANGED) {
+            if (0 > pq_set_guc_locked(self,
+                    "default_transaction_read_only", srv_state_guc[readonly],
+                    &pgres, &error, &_save)) {
+                goto endlock;
+            }
         }
-
-        if ((ret = pq_set_guc_locked(self,
-                "default_transaction_isolation", isolevel->name,
-                &pgres, &error, &_save))) {
-            goto endlock;
+        if (deferrable != SRV_STATE_UNCHANGED) {
+            if (0 > pq_set_guc_locked(self,
+                    "default_transaction_deferrable", srv_state_guc[deferrable],
+                    &pgres, &error, &_save)) {
+                goto endlock;
+            }
+        }
+    }
+    else if (self->autocommit) {
+        /* we are moving from autocommit to not autocommit, so revert the
+         * characteristics to defaults to let BEGIN do its work */
+        if (self->isolevel != ISOLATION_LEVEL_DEFAULT) {
+            if (0 > pq_set_guc_locked(self,
+                    "default_transaction_isolation", "default",
+                    &pgres, &error, &_save)) {
+                goto endlock;
+            }
+        }
+        if (self->readonly != STATE_DEFAULT) {
+            if (0 > pq_set_guc_locked(self,
+                    "default_transaction_read_only", "default",
+                    &pgres, &error, &_save)) {
+                goto endlock;
+            }
+        }
+        if (self->server_version >= 90100 && self->deferrable != STATE_DEFAULT) {
+            if (0 > pq_set_guc_locked(self,
+                    "default_transaction_deferrable", "default",
+                    &pgres, &error, &_save)) {
+                goto endlock;
+            }
         }
-        self->autocommit = 0;
     }
 
-    Dprintf("conn_switch_isolation_level: switched to level %d", level);
+    if (autocommit != SRV_STATE_UNCHANGED) {
+        self->autocommit = autocommit;
+    }
+    if (isolevel != SRV_STATE_UNCHANGED) {
+        self->isolevel = isolevel;
+    }
+    if (readonly != SRV_STATE_UNCHANGED) {
+        self->readonly = readonly;
+    }
+    if (deferrable != SRV_STATE_UNCHANGED) {
+        self->deferrable = deferrable;
+    }
+    rv = 0;
 
 endlock:
     pthread_mutex_unlock(&self->lock);
     Py_END_ALLOW_THREADS;
 
-    if (ret < 0) {
+    if (rv < 0) {
         pq_complete_error(self, &pgres, &error);
+        goto exit;
     }
 
-    return ret;
+    Dprintf(
+        "conn_set_session: autocommit %d, isolevel %d, readonly %d, deferrable %d",
+        autocommit, isolevel, readonly, deferrable);
+
+
+exit:
+    return rv;
 }
 
 
 /* conn_set_client_encoding - switch client encoding on connection */
 
 RAISES_NEG int
-conn_set_client_encoding(connectionObject *self, const char *enc)
+conn_set_client_encoding(connectionObject *self, const char *pgenc)
 {
     PGresult *pgres = NULL;
     char *error = NULL;
     int res = -1;
-    char *codec = NULL;
     char *clean_enc = NULL;
 
+    /* We must know what python encoding this encoding is. */
+    if (0 > clear_encoding_name(pgenc, &clean_enc)) { goto exit; }
+
     /* If the current encoding is equal to the requested one we don't
        issue any query to the backend */
-    if (strcmp(self->encoding, enc) == 0) return 0;
-
-    /* We must know what python codec this encoding is. */
-    if (0 > clear_encoding_name(enc, &clean_enc)) { goto exit; }
-    if (0 > conn_encoding_to_codec(clean_enc, &codec)) { goto exit; }
+    if (strcmp(self->encoding, clean_enc) == 0) return 0;
 
     Py_BEGIN_ALLOW_THREADS;
     pthread_mutex_lock(&self->lock);
@@ -1170,35 +1329,21 @@ conn_set_client_encoding(connectionObject *self, const char *enc)
         goto endlock;
     }
 
-    /* no error, we can proceeed and store the new encoding */
-    {
-        char *tmp = self->encoding;
-        self->encoding = clean_enc;
-        PyMem_Free(tmp);
-        clean_enc = NULL;
-    }
-
-    /* Store the python codec too. */
-    {
-        char *tmp = self->codec;
-        self->codec = codec;
-        PyMem_Free(tmp);
-        codec = NULL;
-    }
-
-    Dprintf("conn_set_client_encoding: set encoding to %s (codec: %s)",
-            self->encoding, self->codec);
-
 endlock:
     pthread_mutex_unlock(&self->lock);
     Py_END_ALLOW_THREADS;
 
-    if (res < 0)
+    if (res < 0) {
         pq_complete_error(self, &pgres, &error);
+        goto exit;
+    }
+
+    res = conn_store_encoding(self, pgenc);
+
+    Dprintf("conn_set_client_encoding: encoding set to %s", self->encoding);
 
 exit:
     PyMem_Free(clean_enc);
-    PyMem_Free(codec);
 
     return res;
 }
@@ -1213,7 +1358,7 @@ exit:
  * until PREPARE. */
 
 RAISES_NEG int
-conn_tpc_begin(connectionObject *self, XidObject *xid)
+conn_tpc_begin(connectionObject *self, xidObject *xid)
 {
     PGresult *pgres = NULL;
     char *error = NULL;
@@ -1247,7 +1392,7 @@ conn_tpc_begin(connectionObject *self, XidObject *xid)
  * for many commands and for recovered transactions. */
 
 RAISES_NEG int
-conn_tpc_command(connectionObject *self, const char *cmd, XidObject *xid)
+conn_tpc_command(connectionObject *self, const char *cmd, xidObject *xid)
 {
     PGresult *pgres = NULL;
     char *error = NULL;
