@@ -166,7 +166,7 @@ typecast_array_tokenize(const char *str, Py_ssize_t strlength,
     return res;
 }
 
-static int
+RAISES_NEG static int
 typecast_array_scan(const char *str, Py_ssize_t strlength,
                     PyObject *curs, PyObject *base, PyObject *array)
 {
@@ -199,7 +199,7 @@ typecast_array_scan(const char *str, Py_ssize_t strlength,
 
             /* before anything else we free the memory */
             if (state == ASCAN_QUOTED) PyMem_Free(token);
-            if (obj == NULL) return 0;
+            if (obj == NULL) return -1;
 
             PyList_Append(array, obj);
             Py_DECREF(obj);
@@ -207,33 +207,37 @@ typecast_array_scan(const char *str, Py_ssize_t strlength,
 
         else if (state == ASCAN_BEGIN) {
             PyObject *sub = PyList_New(0);
-            if (sub == NULL) return 0;
+            if (sub == NULL) return -1;
 
             PyList_Append(array, sub);
             Py_DECREF(sub);
 
-            if (stack_index == MAX_DIMENSIONS)
-                return 0;
+            if (stack_index == MAX_DIMENSIONS) {
+                PyErr_SetString(DataError, "excessive array dimensions");
+                return -1;
+            }
 
             stack[stack_index++] = array;
             array = sub;
         }
 
         else if (state == ASCAN_ERROR) {
-            return 0;
+            return -1;
         }
 
         else if (state == ASCAN_END) {
-            if (--stack_index < 0)
-                return 0;
-            array = stack[stack_index];
+            if (stack_index == 0) {
+                PyErr_SetString(DataError, "unbalanced braces in array");
+                return -1;
+            }
+            array = stack[--stack_index];
         }
 
         else if (state ==  ASCAN_EOF)
             break;
     }
 
-    return 1;
+    return 0;
 }
 
 
@@ -249,23 +253,26 @@ typecast_GENERIC_ARRAY_cast(const char *str, Py_ssize_t len, PyObject *curs)
     Dprintf("typecast_GENERIC_ARRAY_cast: str = '%s',"
             " len = " FORMAT_CODE_PY_SSIZE_T, str, len);
 
-    if (str == NULL) {Py_INCREF(Py_None); return Py_None;}
+    if (str == NULL) { Py_RETURN_NONE; }
     if (str[0] == '[')
         typecast_array_cleanup(&str, &len);
     if (str[0] != '{') {
-        PyErr_SetString(Error, "array does not start with '{'");
+        PyErr_SetString(DataError, "array does not start with '{'");
+        return NULL;
+    }
+    if (str[1] == '\0') {
+        PyErr_SetString(DataError, "malformed array: '{'");
         return NULL;
     }
 
     Dprintf("typecast_GENERIC_ARRAY_cast: str = '%s',"
             " len = " FORMAT_CODE_PY_SSIZE_T, str, len);
 
-    obj = PyList_New(0);
+    if (!(obj = PyList_New(0))) { return NULL; }
 
     /* scan the array skipping the first level of {} */
-    if (typecast_array_scan(&str[1], len-2, curs, base, obj) == 0) {
-        Py_DECREF(obj);
-        obj = NULL;
+    if (typecast_array_scan(&str[1], len-2, curs, base, obj) < 0) {
+        Py_CLEAR(obj);
     }
 
     return obj;
@@ -281,6 +288,7 @@ typecast_GENERIC_ARRAY_cast(const char *str, Py_ssize_t len, PyObject *curs)
 #define typecast_UNICODEARRAY_cast typecast_GENERIC_ARRAY_cast
 #define typecast_BOOLEANARRAY_cast typecast_GENERIC_ARRAY_cast
 #define typecast_DATETIMEARRAY_cast typecast_GENERIC_ARRAY_cast
+#define typecast_DATETIMETZARRAY_cast typecast_GENERIC_ARRAY_cast
 #define typecast_DATEARRAY_cast typecast_GENERIC_ARRAY_cast
 #define typecast_TIMEARRAY_cast typecast_GENERIC_ARRAY_cast
 #define typecast_INTERVALARRAY_cast typecast_GENERIC_ARRAY_cast
