@@ -246,6 +246,13 @@ class ConnectionTests(ConnectingTestCase):
             else:
                 del os.environ['PGCLIENTENCODING']
 
+    def test_connect_no_string(self):
+        class MyString(str):
+            pass
+
+        conn = psycopg2.connect(MyString(dsn))
+        conn.close()
+
     def test_weakref(self):
         from weakref import ref
         import gc
@@ -401,6 +408,13 @@ class ParseDsnTestCase(ConnectingTestCase):
         self.assertRaises(TypeError, ext.parse_dsn, None)
         self.assertRaises(TypeError, ext.parse_dsn, 42)
 
+    def test_str_subclass(self):
+        class MyString(str):
+            pass
+
+        res = ext.parse_dsn(MyString("dbname=test"))
+        self.assertEqual(res, {'dbname': 'test'})
+
 
 class MakeDsnTestCase(ConnectingTestCase):
     def test_empty_arguments(self):
@@ -790,6 +804,14 @@ class IsolationLevelsTestCase(ConnectingTestCase):
         self.assertRaises(ValueError, setattr, self.conn, 'isolation_level', 5)
         self.assertRaises(ValueError, setattr, self.conn, 'isolation_level', 'bah')
 
+    def test_attribs_segfault(self):
+        # bug #790
+        for i in range(10000):
+            self.conn.autocommit
+            self.conn.readonly
+            self.conn.deferrable
+            self.conn.isolation_level
+
 
 class ConnectionTwoPhaseTests(ConnectingTestCase):
     def setUp(self):
