@@ -28,7 +28,7 @@ from __future__ import with_statement
 import psycopg2
 import psycopg2.extensions as ext
 
-from testutils import unittest, ConnectingTestCase
+from testutils import unittest, ConnectingTestCase, skip_before_postgres
 
 
 class WithTestCase(ConnectingTestCase):
@@ -217,6 +217,11 @@ class WithCursorTestCase(WithTestCase):
         else:
             self.fail("where is my exception?")
 
+    @skip_before_postgres(8, 2)
+    def test_named_with_noop(self):
+        with self.conn.cursor('named') as cur:
+            pass
+
 
 def test_suite():
     return unittest.TestLoader().loadTestsFromName(__name__)
